#
#  IBM Confidential
#  OCO Source Materials
#  Rational ClearCase
#  (C) Copyright IBM Corp. 1996, 2003
#
#  The source code for this program is not published or otherwise
#  divested of its trade secrets, irrespective of what has been
#  deposited with the U.S. Copyright Office.
#
### ------------------------------------------------------------------
### | Copyright Unpublished, Atria Software, Inc.  All Rights        |
### | Reserved.  This software contains proprietary and confidential |
### | information of Atria and its suppliers.  Use, disclosure or    |
### | reproduction is prohibited without the prior express written   |
### | consent of Atria.                                              |
### ------------------------------------------------------------------
###

if (-d "C:/") {
    $NT=1;
    $ATRIA_PERL='C:/atria/bin/ccperl';
} else {
    $NT=0;
    $ATRIA_PERL='/usr/atria/bin/Perl';
}
$usage = "usage: $ATRIA_PERL ck_all_tfd_for_nulls.pl vob-stg-pname\n";

# Walk the VOB's source pools and check tfd containers for NULLs.
#
# NOTE: this version is exclusively driven by V2.1 container names
# hence it is only accurate when run on setup V3 VOBs.

# include the source for these below just so this script is 'complete' 
# (since we don't distribute the std perl libraries with ccase)
# if ($NT) {
#     require "find.nt";
# } else
# {
#     require "find.pl";
# }
# require "ctime.pl";

# Traverse desired filesystems

$| = 1; # unbuffered output
$num_ok_conts = 0;
$num_problem_conts = 0;
$vob_s_path = 's';

die($usage) unless $ARGV[0];
$vob_stg = $ARGV[0];
chdir($vob_stg) || die "Can't chdir to VOB storage directory $vob_stg\n";
print "Processing tfd containers under VOB storage dir $vob_stg/s\n";
print "Started at: ", &ctime(time);

if ($NT) {
    &find_nt($vob_s_path);
} else {
    &find($vob_s_path);
}

print "\n\nCompleted at: ", &ctime(time);
print "======= tfd container summary:\n";
print "$num_ok_conts OK\n";
print "$num_problem_conts with ERRORS\n";

exit($num_problem_conts);

# examine files that have names of the form "0-.*"
# (i.e., text_file_delta containers).
#
sub wanted {
    (($dev,$ino,$mode,$nlink,$uid,$gid) = lstat($_)) &&
    -f _ &&
    /^0-.*$/ &&
    print('.') &&
    (&tfd_has_nulls($name, $_) ? $num_problem_conts++ : $num_ok_conts++);
}

# tfd_has_nulls (pname, leafname)
# check (and report) if the leafname file has any nulls in it.
# Also report problems if container is empty or doesn't look like tfd container.
# return TRUE if container has any problems.
#
sub tfd_has_nulls {
    local($pname, $cntr) = @_;
    local($buf, $bcnt, $problems, $fh, @stat);

    # Report a problem if the container is empty.
    @stat = stat($cntr);
    if ($stat[7] <= 0) {
	print "\nERROR: tfd container has 0 size: $pname\n";
	return 1;
    }

    $fh = TFD_HAS_NULLS_FH;
    open($fh, $cntr) || (print "\nERROR: Can't open $pname\n" && return 1);

    # Report if 1st line doesn't have TFD 'control record'
    if ($bcnt = read ($fh, $buf, 25)) {
	if ($buf !~ /^\^S db1 /) {
	    print "\nERROR: missing tfd control record at byte 0 in $pname\n";
	    $problems++;
	}
    }

    # Report any blocks of NULLs
    $problems += &fh_file_has_nulls ($fh, $pname, $cntr);

    close($fh);
    $problems > 0;  # return TRUE if any problems.
}

# file_has_nulls (pname, leafname)
# check (and report) if the leafname file has any nulls in it.
# return TRUE if container has any problems.
#
sub file_has_nulls {
    local($pname, $file) = @_;
    local($problems, $fh);

    $fh = FILE_HAS_NULLS_FH;
    open($fh, $file) || (print "\nERROR: Can't open $pname\n" && return 1);

    # Report any blocks of NULLs
    $problems = &fh_file_has_nulls ($fh, $pname, $file);

    close($fh);

    $problems > 0;  # return TRUE if any problems.
}

# fh_file_has_nulls (fh, pname, leafname)
# check (and report) if the file associated with filehandle 'fh'
# has any nulls in it.
# return TRUE if container has any problems.
#
sub fh_file_has_nulls {
    local($fh, $pname, $file) = @_;
    local($buf, $bcnt, $bufsize, $bufbaseindx);
    local($num_null_blocks, $problems);
    local($pos, $first_null_indx, $last_null_indx);

    # reset things 
    seek($fh, 0, 0);
    binmode($fh);

    $bufsize=16384;

    $bufbaseindx=0;
    $last_null_indx = -2;  # must be < -1 to ensure 1st NULL (at loc 0)
			   # looks like a new block
    while ($bcnt = read ($fh, $buf, $bufsize)) {
	$dbg > 1 && print "processing $bcnt bytes at $bufbaseindx\n";

	# Track / report blocks of NULLs (deal with blocks spanning buffers)
	$pos = $[;
	while (($pos = index($buf, "\000", $pos)) >= $[) {
	    # Found a NULL.
	    if ($pos - 1 + $bufbaseindx == $last_null_indx) {
		# Consecutive, just keep going.
		$last_null_indx++;
	    } else {
		# Begin new block (report end of previous block if any)
		print $last_null_indx - $first_null_indx + 1,
		      " consecutive NULL bytes\n" if ($last_null_indx >= 0);

		# begin new block
		$first_null_indx = $pos + $bufbaseindx;
		$last_null_indx = $first_null_indx;
		print "\nERROR: found NULL at byte $first_null_indx in $pname\n";
		$problems++;
		$num_null_blocks++;
	    }
	    $pos++;
	}
	$bufbaseindx += $bcnt;
    }
    # Report the end of the last NULL block (if any)
    if ($last_null_indx >= 0) {
        if ($first_null_indx == 0 && $last_null_indx + 1 == $bufbaseindx) {
            print $last_null_indx - $first_null_indx + 1,
                " consecutive NULL bytes - FILE IS ALL NULLS\n";
        } else {
            print $last_null_indx - $first_null_indx + 1,
                " consecutive NULL bytes\n";
        }
    }

    $problems > 0;  # return TRUE if any problems.
}

####
#### directly included source from the perl libraries
####

;# ctime.pl is a simple Perl emulation for the well known ctime(3C) function.
;#
;# Waldemar Kebsch, Federal Republic of Germany, November 1988
;# kebsch.pad@nixpbe.UUCP
;# Modified March 1990, Feb 1991 to properly handle timezones
;#  $RCSfile: ctime.pl,v $$Revision: 4.0.1.1 $$Date: 92/06/08 13:38:06 $
;#   Marion Hakanson (hakanson@cse.ogi.edu)
;#   Oregon Graduate Institute of Science and Technology
;#
;# usage:
;#
;#     #include <ctime.pl>          # see the -P and -I option in perl.man
;#     $Date = &ctime(time);

CONFIG: {
    package ctime;

    @DoW = ('Sun','Mon','Tue','Wed','Thu','Fri','Sat');
    @MoY = ('Jan','Feb','Mar','Apr','May','Jun',
	    'Jul','Aug','Sep','Oct','Nov','Dec');
}

sub ctime {
    package ctime;

    local($time) = @_;
    local($[) = 0;
    local($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst);

    # Determine what time zone is in effect.
    # Use GMT if TZ is defined as null, local time if TZ undefined.
    # There's no portable way to find the system default timezone.

    $TZ = defined($ENV{'TZ'}) ? ( $ENV{'TZ'} ? $ENV{'TZ'} : 'GMT' ) : '';
    ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) =
        ($TZ eq 'GMT') ? gmtime($time) : localtime($time);

    # Hack to deal with 'PST8PDT' format of TZ
    # Note that this can't deal with all the esoteric forms, but it
    # does recognize the most common: [:]STDoff[DST[off][,rule]]

    if($TZ=~/^([^:\d+\-,]{3,})([+-]?\d{1,2}(:\d{1,2}){0,2})([^\d+\-,]{3,})?/){
        $TZ = $isdst ? $4 : $1;
    }
    $TZ .= ' ' unless $TZ eq '';

    $year += ($year < 70) ? 2000 : 1900;
    sprintf("%s %s %2d %2d:%02d:%02d %s%4d\n",
      $DoW[$wday], $MoY[$mon], $mday, $hour, $min, $sec, $TZ, $year);
}


# ================ NT version
# Usage:
#	require "find.pl";
#
#	&find('/foo','/bar');
#
#	sub wanted { ... }
#		where wanted does whatever you want.  $dir contains the
#		current directory name, and $_ the current filename within
#		that directory.  $name contains "$dir/$_".  You are cd'ed
#		to $dir when the function is called.  The function may
#		set $prune to prune the tree.
#
# This library is primarily for find2perl, which, when fed
#
#   find2perl / -name .nfs\* -mtime +7 -exec rm -f {} \; -o -fstype nfs -prune
#
# spits out something like this
#
#	sub wanted {
#	    /^\.nfs.*$/ &&
#	    (($dev,$ino,$mode,$nlink,$uid,$gid) = lstat($_)) &&
#	    int(-M _) > 7 &&
#	    unlink($_)
#	    ||
#	    ($nlink || (($dev,$ino,$mode,$nlink,$uid,$gid) = lstat($_))) &&
#	    $dev < 0 &&
#	    ($prune = 1);
#	}

sub find_nt {
    chop($cwd = `cd`);
    foreach $topdir (@_) {
	(($topdev,$topino,$topmode,$topnlink) = stat($topdir))
	  || (warn("Can't stat $topdir: $!\n"), next);
	if (-d _) {
	    if (chdir($topdir)) {
		($dir,$_) = ($topdir,'.');
		$name = $topdir;
		&wanted;
		$topdir =~ s,[/\\]$,, ;
		&finddir_nt($topdir);
	    }
	    else {
		warn "Can't cd to $topdir: $!\n";
	    }
	}
	else {
	    unless (($dir,$_) = $topdir =~ m#^(.*[/\\])(.*)$#) {
		($dir,$_) = ('.', $topdir);
	    }
	    $name = $topdir;
	    chdir $dir && &wanted;
	}
	chdir $cwd;
    }
}

sub finddir_nt {
    local($dir) = @_;
    local($dev,$ino,$mode,$subcount);
    local($name);

    # Get the list of files in the current directory.

    opendir(DIR,'.') || (warn "Can't open $dir: $!\n", return);
    local(@filenames) = readdir(DIR);
    closedir(DIR);

    for (@filenames) {
	next if $_ eq '.';
	next if $_ eq '..';
	$name = "$dir/$_";
	$nlink = 0;
	&wanted;
	if (-d $_) {    
	    # It really is a directory, so do it recursively.
	    
	    if (!$prune && chdir $_) {
		&finddir_nt($name);
		chdir '..';
	    }
	}
    }
}

# ================ UNIX version
# Usage:
#	require "find.pl";
#
#	&find('/foo','/bar');
#
#	sub wanted { ... }
#		where wanted does whatever you want.  $dir contains the
#		current directory name, and $_ the current filename within
#		that directory.  $name contains "$dir/$_".  You are cd'ed
#		to $dir when the function is called.  The function may
#		set $prune to prune the tree.
#
# This library is primarily for find2perl, which, when fed
#
#   find2perl / -name .nfs\* -mtime +7 -exec rm -f {} \; -o -fstype nfs -prune
#
# spits out something like this
#
#	sub wanted {
#	    /^\.nfs.*$/ &&
#	    (($dev,$ino,$mode,$nlink,$uid,$gid) = lstat($_)) &&
#	    int(-M _) > 7 &&
#	    unlink($_)
#	    ||
#	    ($nlink || (($dev,$ino,$mode,$nlink,$uid,$gid) = lstat($_))) &&
#	    $dev < 0 &&
#	    ($prune = 1);
#	}

sub find {
    chop($cwd = `pwd`);
    foreach $topdir (@_) {
	(($topdev,$topino,$topmode,$topnlink) = stat($topdir))
	  || (warn("Can't stat $topdir: $!\n"), next);
	if (-d _) {
	    if (chdir($topdir)) {
		($dir,$_) = ($topdir,'.');
		$name = $topdir;
		&wanted;
		$topdir =~ s,/$,, ;
		&finddir($topdir,$topnlink);
	    }
	    else {
		warn "Can't cd to $topdir: $!\n";
	    }
	}
	else {
	    unless (($dir,$_) = $topdir =~ m#^(.*/)(.*)$#) {
		($dir,$_) = ('.', $topdir);
	    }
	    $name = $topdir;
	    chdir $dir && &wanted;
	}
	chdir $cwd;
    }
}

sub finddir {
    local($dir,$nlink) = @_;
    local($dev,$ino,$mode,$subcount);
    local($name);

    # Get the list of files in the current directory.

    opendir(DIR,'.') || (warn "Can't open $dir: $!\n", return);
    local(@filenames) = readdir(DIR);
    closedir(DIR);

    if ($nlink == 2) {        # This dir has no subdirectories.
	for (@filenames) {
	    next if $_ eq '.';
	    next if $_ eq '..';
	    $name = "$dir/$_";
	    $nlink = 0;
	    &wanted;
	}
    }
    else {                    # This dir has subdirectories.
	$subcount = $nlink - 2;
	for (@filenames) {
	    next if $_ eq '.';
	    next if $_ eq '..';
	    $nlink = $prune = 0;
	    $name = "$dir/$_";
	    &wanted;
	    if ($subcount > 0) {    # Seen all the subdirs?

		# Get link count and check for directoriness.

		($dev,$ino,$mode,$nlink) = lstat($_) unless $nlink;
		
		if (-d _) {

		    # It really is a directory, so do it recursively.

		    if (!$prune && chdir $_) {
			&finddir($name,$nlink);
			chdir '..';
		    }
		    --$subcount;
		}
	    }
	}
    }
}


